
--[[

   Written by Francois Fleuret (francois@fleuret.org)

   This is free and unencumbered software released into the public
   domain.

   Anyone is free to copy, modify, publish, use, compile, sell, or
   distribute this software, either in source code form or as a
   compiled binary, for any purpose, commercial or non-commercial, and
   by any means.

   In jurisdictions that recognize copyright laws, the author or
   authors of this software dedicate any and all copyright interest in
   the software to the public domain. We make this dedication for the
   benefit of the public at large and to the detriment of our heirs
   and successors. We intend this dedication to be an overt act of
   relinquishment in perpetuity of all present and future rights to
   this software under copyright law.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   For more information, please refer to <http://unlicense.org/>

]]--

require 'torch'
require 'nn'
require 'sys'

profiler = {}

function profiler.decorate(model, functionsToDecorate)

   local functionsToDecorate = functionsToDecorate or
      {
         'updateOutput',
         'backward'
      }

   for _, name in pairs(functionsToDecorate) do
      model.accTime = {}

      local nameOrig = name .. '__orig'

      -- We decorate the class and not the object, otherwise we cannot
      -- save models anymore.

      if rawget(model, name) then
         error('We decorate the class, not the objects, and there is a ' .. name .. ' in ' .. model)
      end

      local toDecorate = getmetatable(model)

      if toDecorate[name] and not toDecorate[nameOrig] then
         toDecorate[nameOrig] = toDecorate[name]
         toDecorate[name] = function(self, ...)
            local startTime = sys.clock()
            local result = { self[nameOrig](self, unpack({...})) }
            local endTime = sys.clock()
            self.accTime[name] = (self.accTime[name] or 0) + endTime - startTime
            return unpack(result)
         end
      end

   end

   if torch.isTypeOf(model, nn.Container) then
      for _, m in ipairs(model.modules) do
         profiler.decorate(m, functionsToDecorate)
      end
   end

end

function profiler.print(model, nbSamples, totalTime, indent)
   local indent = indent or ''

   print(string.format('%s* %s', indent, model.__typename))

   for l, t in pairs(model.accTime) do
      local s = string.format('%s  %s %.02fs', indent, l, t)
      if totalTime then
         s = s .. string.format(' [%.02f%%]', 100 * t / totalTime)
      end
      if nbSamples then
         s = s .. string.format(' (%.01fmus/sample)', 1e6 * t / nbSamples)
      end
      print(s)
   end

   print()

   if torch.isTypeOf(model, nn.Container) then
      for _, m in ipairs(model.modules) do
         profiler.print(m, nbSamples, totalTime, indent .. '  ')
      end
   end
end

return profiler
